-- ForthEC parser
-- /Mic, 2004


include file.e
include get.e
include machine.e
include wildcard.e


global constant WORDS = 
{
"!",		-- (n adr --)
"!r",		-- (adr n --)
"!r+",		-- (adr n -- adr) store n in adr and increment adr by 4
"\"",
"\".",
"#align",	-- (-- n)	alignment granularity
"$\"",		-- 
"\'",
"(",		-- (--)		begin comment	
")",		-- (--)		end comment
"*",		-- (n1 n2 -- n3)
"+",		-- (n1 n2 -- n3)
"+!",		-- (n adr --) 	add n to the value stored att adr and store the new value
"+loop",
",",
"-",		-- (n1 n2 --)
"--",
"-rot",
".",		-- (n --)
".\"",
".(",
".s",
"/",		-- (n1 n2 -- n3)
"/c",		-- (-- n)	size of character (byte)
"/c*",
"/f",		-- (-- n)	size of floating point number in bytes
"/f*",
"/l",
"/l*",
"/mod",		-- (n1 n2 -- n3 n4)
"/n",
"/n*",
"/w",
"/w*",
"0<",
"0<=",
"0<>",
"0=",		-- (n1 n2 -- f)
"0>",
"0>=",
"1+",		-- (n1 -- n2)
"1-",		-- (n1 -- n2)
"16-bit",
"2*",		-- (n1 -- n2)
"2+",		-- (n1 -- n2)
"2/",		-- (n1 -- n2)
"2drop",
"2dup",
"2over",
"32-bit",
":",
":noname",
":w",
";",
";@",
";w",
"<",
"<<",
"<=",
"<>",
"<w@",
"=",
">",
">=",
">>",
">>a",
"?",
"?do",
"?dup",
"?leave",
"?stack",
"@",
"\\",
"a!",
"a@",
"abort",
"abs",		-- (n1 -- n2)
"alias",
"align",
"aligned",
"allot",
"and",
"ascii",	-- (n1 -- n2)
"begin",
"between",
"bl",
"bounds",
"branch",
"bye",
"c!",
"c,",
"c:es!",
"c:es@",
"c@",
"ca+",
"call",
"calli",
"canonical",
"case",
"cell",
"cell+",
"cells",
"cmove",
"code",
"constant",
"convert",
"cr",		-- (-- char)
"cs@",
"decimal",	-- (--)
"default",
"defer",
"depth",
"digit",
"do",
"drop",
"dup",
"else",
"emit",
"emms",
"end-string-array",
"endcase",
"endof",
"entry",
"erase",
"es!",
"exit",		-- (--)
"expect",	-- (adr n --)
"extern",
"f!",		-- (fp adr --)
"f!s",		-- (fp adr --)
"f*",
"f**",
"f+",		-- (fp1 fp2 -- fp3)
"f-",
"f.",		-- (fp --)		display fp
"f.s",
"f/",		-- (fp1 fp2 -- fp3)
"f0=",
"f0<>",		
"f<",
"f<>",
"f=",
"f>",
"f@",		-- (adr -- fp)
"f@s",
"fabs",
"falog",
"false",
"fatan",
"fclose",
"fconstant",
"fcos",
"fcr",
"fdepth",	-- (-- n)
"fdrop",
"fdup",
"fexit",
"ffills",
"fflush",
"fgetc",
"fgetline",
"fgets",
"fill",
"fix",		-- (fp -- l)		convert fp to integer (round towards nearest int)
"float",	-- (l -- fp)		convert integer to float
"flog",
"floor",
"fnegate",
"fopen",
"fover",
"fpop",		-- (fp -- l l)
"fpops",	-- (fp -- l)
"fpush",
"fpushs",	-- (l -- fp)		single precision version of fpush
"fputc",
"fputs",
"frot",
"fseek",
"fsin",
"fsincos",	-- (fp1 -- fp2 fp3)
"fsize",
"fsqrt",	-- (fp1 -- fp2)
"fswap",
"ftan",
"ftell",
"fvariable",
"here",
"hex",		-- (--)
"i",
"if",
"include",
"inp",
"int",		-- (fp -- l)		round towards zero
"interrupt",
"j",
"key",
"l+!",
"l.",
"l0=",
"l<",
"l=",
"l>",
"l@",
"la+",
"land",
"leave",
"loop",
"max",		-- (n1 n2 -- n3)
"min",		-- (n1 n2 -- n3)
"mod",		-- (n1 n2 -- n3)
"move",
"mu/mod",
"n->w",
"na+",
"na+1",
"negate",
"newline",	-- (-- char)
"nip",		-- (n1 n2 -- n2)
"noop",		-- (--)
"not",		-- (n1 -- n2)
"notouch",
"null",		-- (-- 0)
"number?",	-- (adr -- l flag)	
"of",		-- (selector test-value --)
"off",		-- (adr --)
"on",		-- (adr --)		stores true at adr
"or",		-- (n1 n2 -- n3)	logical or
"outp",
"over",		-- (n1 n2 -- n1 n2 n3)	n3 is a copy of n1
"p!",
"p+",
"p-",
"p@",
"p@a",
"pc+",
"pc-",
"pc2pw",
"pdrop",
"pdup",
"pf!",
"pf!a",
"pf@",
"pf@a",
"pick",
"printable?",
"pw*",
"pw+",
"pw-",
"pw2pc",
"pw2pcs",
"pw2pd",
"repeat",
"rot",		-- (n1 n2 n3 -- n2 n3 n1)
"s.",		-- (n --)
"sp!",
"sp0",
"sp@",
"space",	-- (--)
"spaces",	-- (n --)
"string-array",
"swap",		-- (n1 n2 -- n2 n1)
"then",
"to",
"touch",
"true",
"tuck",
"type",
"u.",
"u<",
"u<=",
"u>",
"u>=",
"ul*",
"um*",
"um/mod",
"until",
"value",
"variable",
"w!",
"w,",
"w->l",
"w:es!",
"w@",
"wa+",
"wa+1",
"while",
"within",
"wl@",
"xor",		-- (n1 n2 -- n3)
"z\"",
"{",
"}"
}


global constant
	W_INCLUDE	= find("include",WORDS),
	W_DEFER 	= find("defer",WORDS),
	W_HERE		= find("here",WORDS),

	W_CELL		= find("cell",WORDS),
	W_CELLPLUS	= find("cell+",WORDS),
	W_CELLS		= find("cells",WORDS),
	
	W_DOT		= find(".",WORDS),
	W_DOTS		= find(".s",WORDS),
	W_DISPVAR	= find("?",WORDS),
	W_DOTSTRING	= find(".\"",WORDS),
	W_FDOT		= find("f.",WORDS),
	W_EMIT		= find("emit",WORDS),
	W_ASCII		= find("ascii",WORDS),
	W_BL		= find("bl",WORDS),
	
	W_PUTS		= find("\".",WORDS),
	W_EXPECT	= find("expect",WORDS),
	W_KEY		= find("key",WORDS),
	
	W_CR		= find("cr",WORDS),
	W_NEWLINE	= find("newline",WORDS),
	W_SPACE		= find("space",WORDS),
	W_SPACES	= find("spaces",WORDS),
	
	W_COLON		= find(":",WORDS),
	W_SEMICOLON	= find(";",WORDS),
	W_SEMICOLONPOP	= find(";@",WORDS),
	W_COLONWIN	= find(":w",WORDS),
	W_SEMICOLONWIN	= find(";w",WORDS),
	W_COLONNO	= find(":noname",WORDS),
	W_QUOTE		= find("\'",WORDS),
	W_COMMA		= find(",",WORDS),
	W_CCOMMA	= find("c,",WORDS),
	W_WCOMMA	= find("w,",WORDS),

	W_BOUNDS	= find("bounds",WORDS),
	
	W_BEGINCOMMENT	= find("(",WORDS),
	W_ENDCOMMENT	= find(")",WORDS),
	W_SINGLECOMMENT	= find("\\",WORDS),

	W_PARAMSTART	= find("{",WORDS),
	W_PARAMEND	= find("}",WORDS),
	W_PARAMDELIM	= find("--",WORDS),
	
	W_CONST		= find("constant",WORDS),
	W_VAR		= find("variable",WORDS),
	W_FCONST	= find("fconstant",WORDS),
	W_FVAR		= find("fvariable",WORDS),
	W_EXTERN	= find("extern",WORDS),
	
	W_NULL		= find("null",WORDS),
	W_TRUE		= find("true",WORDS),
	W_FALSE		= find("false",WORDS),
	
	W_STORE		= find("!",WORDS),
	W_CSTORE	= find("c!",WORDS),
	W_FSTORE	= find("f!",WORDS),
	W_FSTORES	= find("f!s",WORDS),
	W_WSTORE	= find("w!",WORDS),
	W_STORE_R	= find("!r",WORDS),
	W_STORE_R_INC	= find("!r+",WORDS),
	W_ERASE		= find("erase",WORDS),
	W_FILL		= find("fill",WORDS),
	W_FFILLS	= find("ffills",WORDS),
	W_CMOVE		= find("cmove",WORDS),
	W_FETCH		= find("@",WORDS),
	W_CFETCH	= find("c@",WORDS),
	W_CSFETCH	= find("cs@",WORDS),
	W_WFETCH	= find("w@",WORDS),
	W_WLOFETCH	= find("wl@",WORDS),
	W_FFETCH	= find("f@",WORDS),
	W_FFETCHS	= find("f@s",WORDS),
	W_ACCFETCH	= find("a@",WORDS),

	W_PSTORE	= find("p!",WORDS),
	W_PFETCH	= find("p@",WORDS),
	W_PWADD		= find("pw+",WORDS),
	W_PWSUB		= find("pw-",WORDS),
	W_PWMUL		= find("pw*",WORDS),
	W_PCADD		= find("pc+",WORDS),
	W_PCSUB		= find("pc-",WORDS),
	W_PC2PW		= find("pc2pw",WORDS),
	W_PW2PC		= find("pw2pc",WORDS),
	W_PW2PCS	= find("pw2pcs",WORDS),
	W_PDUP		= find("pdup",WORDS),
	W_PDROP 	= find("pdrop",WORDS),
	W_EMMS		= find("emms",WORDS),
	
	W_ON		= find("on",WORDS),
	W_OFF		= find("off",WORDS),
	
	W_SPBOT		= find("sp0",WORDS),
	W_SPTOP		= find("sp@",WORDS),
	
	W_FPOP		= find("fpop",WORDS),
	W_FPOPS		= find("fpops",WORDS),
	
	W_CSIZE		= find("/c",WORDS),
	W_FSIZE		= find("/f",WORDS),
	W_NSIZE		= find("/n",WORDS),
	
	W_DUP		= find("dup",WORDS),
	W_DUPNZ		= find("?dup",WORDS),
	W_OVER		= find("over",WORDS),
	W_PICK		= find("pick",WORDS),
	W_NIP		= find("nip",WORDS),
	W_DROP		= find("drop",WORDS),
	W_ROT		= find("rot",WORDS),
	W_ROTINV	= find("-rot",WORDS),
	W_SWAP		= find("swap",WORDS),
	W_TUCK		= find("tuck",WORDS),
	
	W_2DROP		= find("2drop",WORDS),
	W_2DUP		= find("2dup",WORDS),
	W_2OVER		= find("2over",WORDS),

	W_FDUP		= find("fdup",WORDS),
	W_FDROP		= find("fdrop",WORDS),
	W_FOVER		= find("fover",WORDS),
	W_FSWAP		= find("fswap",WORDS),

	W_ADD		= find("+",WORDS),
	W_SUB		= find("-",WORDS),
	W_MUL		= find("*",WORDS),
	W_DIV		= find("/",WORDS),
	W_MOD		= find("mod",WORDS),
	W_DIVMOD	= find("/mod",WORDS),

	W_MIN		= find("min",WORDS),
	W_MAX		= find("max",WORDS),
	W_BETWEEN	= find("between",WORDS),
	W_WITHIN	= find("within",WORDS),
	
	W_FADD		= find("f+",WORDS),
	W_FSUB		= find("f-",WORDS),
	W_FMUL		= find("f*",WORDS),
	W_FDIV		= find("f/",WORDS),

	W_FSIN		= find("fsin",WORDS),
	W_FCOS		= find("fcos",WORDS),

	W_FLOAT		= find("float",WORDS),
	W_FLOOR		= find("floor",WORDS),
	
	W_MUL2		= find("2*",WORDS),
	W_DIV2		= find("2/",WORDS),
	W_DEC		= find("1-",WORDS),
	W_INC		= find("1+",WORDS),

	W_NEG		= find("negate",WORDS),
	W_FNEG		= find("fnegate",WORDS),
	
	W_SHL		= find("<<",WORDS),
	W_SHR		= find(">>",WORDS),
	
	W_AND		= find("and",WORDS),
	W_OR		= find("or",WORDS),
	W_XOR		= find("xor",WORDS),
	W_NOT		= find("not",WORDS),

	W_EQ		= find("=",WORDS),
	W_NE		= find("<>",WORDS),
	W_GE		= find(">=",WORDS),
	W_GT		= find(">",WORDS),
	W_LE		= find("<=",WORDS),
	W_LT		= find("<",WORDS),
	W_EQ0		= find("0=",WORDS),
	W_NE0		= find("0<>",WORDS),
	W_GT0		= find("0>",WORDS),
	W_LT0		= find("0<",WORDS),

	W_FEQ		= find("f=",WORDS),
	W_FNE		= find("f<>",WORDS),
	W_FGT		= find("f>",WORDS),
	W_FLT		= find("f<",WORDS),
	
	W_BEGIN		= find("begin",WORDS),
	W_WHILE		= find("while",WORDS),
	W_REPEAT	= find("repeat",WORDS),
	W_DO		= find("do",WORDS),
	W_LOOP		= find("loop",WORDS),
	W_INCLOOP	= find("+loop",WORDS),
	W_LEAVE		= find("leave",WORDS),
	W_LEAVETRUE	= find("?leave",WORDS),
	W_UNTIL		= find("until",WORDS),
	W_I		= find("i",WORDS),
	W_J		= find("j",WORDS),
	
	W_IF		= find("if",WORDS),
	W_ELSE		= find("else",WORDS),
	W_THEN		= find("then",WORDS),
	
	W_CASE		= find("case",WORDS),
	W_ENDCASE	= find("endcase",WORDS),
	W_OF		= find("of",WORDS),
	W_ENDOF		= find("endof",WORDS),
	W_DEFAULT	= find("default",WORDS),
	
	W_CALL		= find("call",WORDS),
	W_CALLI		= find("calli",WORDS),
	W_BYE		= find("bye",WORDS),
	W_EXIT		= find("exit",WORDS),
	
	W_ABS		= find("abs",WORDS),

	W_ALLOT		= find("allot",WORDS),	
	W_ZSTRING	= find("z\"",WORDS),
	W_DOLLARSTRING	= find("$\"",WORDS),

	W_INTERRUPT	= find("interrupt",WORDS),
	W_INP		= find("inp",WORDS),
	W_OUTP		= find("outp",WORDS),
	W_ACCSTORE	= find("a!",WORDS),
	W_CESFETCH	= find("c:es@",WORDS),
	W_CESSTORE	= find("c:es!",WORDS),
	W_WESSTORE	= find("w:es!",WORDS),
	W_ESSTORE	= find("es!",WORDS),
	
	W_TOUCH		= find("touch",WORDS),
	W_NOTOUCH	= find("notouch",WORDS)


--W_PW2PC

global constant
	HEXNUMBER = -5,
	NUMBER  = -4,
	FNUMBER = -3,
	NEWLINE = -2,
	UNKNOWN = -1,
	NOTHING = 0


global sequence
	constants,
	fconstants,
	tokens

global integer 
	verbose
	
	
integer
	isString


isString = 0
verbose = 0


-- Do a binary search for 'look_for' in 'look_in'.
-- 'look_in' is assumed to be lexically ordered.
global function bsearch(sequence look_for,sequence look_in)
	integer lb,ub,middle,res
	
	lb = 0
	ub = length(look_in)-1
	
	while 1 do
		if ub<lb then
			return 0
		end if

		middle = floor((lb+ub)/2)
		
		res = compare(look_for,look_in[middle+1])
		if res=0 then
			return middle+1
		elsif res<0 then
			ub = middle-1
		else
			lb = middle+1
		end if
	end while
	return 0
end function


--? bsearch("notouch",WORDS)

global procedure ERROR(sequence msg,sequence token)
	puts(1,"ERROR: "&msg)
	if sequence(token[2]) then
		puts(1,": "&token[2])
	end if
	printf(1," (line %d)\nPress any key to abort..",token[3])
	while get_key()=-1 do end while
	abort(0)
end procedure



function get_token(integer fn,integer line)
	integer c,p,ttype
	sequence token,token2,val
	
	token = {}
	ttype = NUMBER
	
	if not isString then
		c = ' '
		while (c=' ' or c='\t' or c=13) do
			c = getc(fn)
			if c = -1 then
				return {NOTHING}
			end if
		end while
	else
		c = getc(fn)
	end if
	
	while (isString and c!='\"') or
	      ((not isString) and (not (c=' ' or c='\t' or c=13))) do
		token &= c

		if c=10 then
			exit
		elsif c>='0' and c<='9' then
		elsif c='.' then
			if ttype=NUMBER and length(token)>1 then
				ttype = FNUMBER
			else
				ttype = UNKNOWN
			end if
		elsif c='e' or c='E' then
			if ttype=HEXNUMBER then
			elsif ttype != FNUMBER and length(token)>1 then
				ttype = UNKNOWN
			end if
		elsif c='x' or c='X' and length(token)=2 and ttype=NUMBER then
			if token[1]='0' then
				ttype = HEXNUMBER
			else
				ttype = UNKNOWN
			end if
		elsif c='-' then
			if length(token)>1 then
				ttype = UNKNOWN
			end if
		elsif (c=' ' or c='\t' or c=13) then
		else
			if ttype=HEXNUMBER and (c>='a' and c<='f') then
				token[length(token)] -= ' '
			elsif ttype=HEXNUMBER and (c>='A' and c<='F') then
			else
				ttype = UNKNOWN
			end if
		end if

		c = getc(fn)
		if c=-1 then exit end if
	end while
	
	if isString then
		token = token & "\""
		return {UNKNOWN,token,line}
	end if
	
	if not length(token) then
		return {NOTHING}
	
	elsif equal(token,{10}) then
		return {NEWLINE,token,line}
	
	else
		token2 = lower(token)
		p = bsearch(token2,WORDS)
		if p then
			--if p = W_NOTOUCH then ? p end if
			return {p,token,line}
		end if
	end if

	if ttype = NUMBER then
		val = value(token)
		return {NUMBER,val[2],line}
	
	elsif ttype = FNUMBER then
		val = value(token)
		return {FNUMBER,val[2],line}
	
	elsif ttype=HEXNUMBER then
		val = value("#"&token[3..length(token)])
		return {NUMBER,val[2],line}
	end if
	
	
	return {UNKNOWN,token,line}
end function



global function parse(sequence fname)
	integer fn,n,p,line,touch
	sequence s,t,token,asm
	
	fn = open(fname,"rb")
	if fn=-1 then
		ERROR("Unable to open file",{0,fname,1})
	end if
	
	if verbose then
		puts(1,"Parsing "&fname&"\n")
	end if
	
	line = 1
	touch = 1
	
	-- First stage: tokenize input stream	
	while 1 do
		token =	get_token(fn,line)
		
		if token[1]=NOTHING then
			exit
		elsif token[1]=NEWLINE then
			if not touch then
				tokens = append(tokens,token)
			end if
			line += 1
		elsif token[1] = W_BEGINCOMMENT then		-- start of comment
			while token[1]!=W_ENDCOMMENT and token[1]!=NOTHING do
				token = get_token(fn,line)
				if token[1]=NEWLINE then
					line += 1
				end if
			end while
		elsif token[1] = W_SINGLECOMMENT then
			while not (token[1]=NEWLINE or token[1]=NOTHING) do
				token = get_token(fn,line)
			end while
			line += 1
		elsif token[1]=W_ZSTRING or
		      token[1]=W_DOTSTRING or
		      token[1]=W_DOLLARSTRING then
			tokens = append(tokens,token)
			isString = 1
			tokens = append(tokens,get_token(fn,line))
			isString = 0
		elsif token[1]=W_TOUCH then
			tokens = append(tokens,token)
			touch = 1
		elsif token[1]=W_NOTOUCH then
			tokens = append(tokens,token)
			touch = 0
		elsif token[1]=W_INCLUDE and touch then
			token = get_token(fn,line)
			if parse(token[2]) then
			end if
		else
			tokens = append(tokens,token)
		end if
	end while
	
	close(fn)
	return 1
end function
